# Light Fogging Correction tool for ImageJ
# Ver. 0.09 version
# write out R-G dif value
# make some processes sub routines
# add input Power for Mixing Tuning
# (c)2021-22 Ohnishi, Yasuo
from ij import IJ, ImagePlus, CompositeImage
from ij import WindowManager as WM
from ij.io import DirectoryChooser, OpenDialog
from ij.gui import GenericDialog
from ij.plugin import ChannelSplitter, Duplicator
from ij.plugin import RGBStackMerge, RGBStackConverter
from ij.process import ImageConverter, ImageProcessor
from loci.plugins import BF
import math
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')
import datetime
from java.awt.event import ItemListener, TextListener
from java.lang import Integer
import sys

def convValue(Value, Mode):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if Mode == "16-bit":
        Value = Value * 256
    elif Mode == "32-bit":
        Value = Value * 256 * 256
    else:
        Value = Value
    return Value

####### calculating R-G mean difference in Sampling Region  ############
def calcMean(imp0, imp1, xStart, xEnd, nYDivide, yStart, yEnd, folder):
#   imp0: R image / imp1: G image / xStart & xEnd: Starting & Ending Point (x) of ROI
#   nYDivide:  number of dividing zones in Y
    impTmp = []
    impTmp.append(imp0)
    impTmp.append(imp1)
    nYDivide = int(nYDivide) # number of dividing zones in Y
    width = imp0.getWidth() # get image width
    height = imp0.getHeight() # get image height
    pixelNum = width*height # get total pixel number
    print ("W x H (imp0)", str(width), str(height))
    if yStart == 0 and yEnd == 0:
        height2 = height
    else:
        height2 = yEnd - yStart
    zoneBreadth = height2 // nYDivide # divided Y zone height
    tmpsumDiff = [0] * nYDivide # sum of R-G differance in a zone
    n_tmpsumDiff = [0] * nYDivide # sum of pixel number in a zone
    mean2 = [0.0] * nYDivide # mean R-G differance in a zone

    for x in range(xStart, xEnd): # processing in each X row
#       ----   
        for yz in range(nYDivide): # processing in each Y zone
            for yj in range(yStart + yz * zoneBreadth, yStart + (yz + 1) * zoneBreadth - 1): 
                Px0 = impTmp[0].getPixel(x, yj) # get R pixel Value
                Px1 = impTmp[1].getPixel(x, yj) # get G pixel Value
                tmpsumDiff[yz] = tmpsumDiff[yz] + (Px0[0] - Px1[0]) # summate R-G difference in each Y Zone
                n_tmpsumDiff[yz] = n_tmpsumDiff[yz] + 1 # count pixel number in each Y Zone
#       ----   
#    for inspecting contents of mean2
#a    writing_file1 = open(folder + 'mean2.txt','w')
    for yz in range(nYDivide): # calculate mean R-G difference in each Y Zone
        mean2[yz] = float(tmpsumDiff[yz]) / float(n_tmpsumDiff[yz])
#a        writing_file1.write(str(yz) + "," + str(mean2[yz]) + "\n")
#       ----   
#a    writing_file1.close()
    return mean2 # Return value: mean[n of blocks] R-G difference in Y Zones
#---------------
#### Getting R-G value difference in each columns in ROI
def get_RG_v_diff(impR, impG, xStart, xEnd, yStart, yEnd, height2, folder, nDivide):
    tmpMxDiffRG = [[0.0 for yj in range(height2)] for xi in range(xEnd - xStart)]
    timenow = datetime.datetime.now()
    print("Start Read tmpMxDiffRG:", timenow)
    for x in range(xStart, xEnd): # process in each x col
        for y in range(yStart, yStart + height2):
#                   get Red image pixel value
            PxR = impR.getPixel(x, y) 
#                   get Green image pixel value
            PxG = impG.getPixel(x, y) 
            tmpMxDiffRG[x - xStart][y - yStart] = PxR[0] - PxG[0]
                 # save R-G pixel value difference to tmpMxDiffRG
    """
    # for inspecting contents of tmpMxDiffRG
    tmpMxDiffRG_writeout(tmpMxDiffRG, folder, nDivide)
    """
    return tmpMxDiffRG
     # tmpMxDiffRG[width pixes] x [n of blocks]: R-G difference value in ROI 
def tmpMxDiffRG_writeout(tmpMxDiffRG, folder, nDivide):
    writing_file2 = open(folder + 'tmpMxDiffRG.txt','w')
    xlen = len(tmpMxDiffRG)
    writing_file2.write("xlen: " + str(xlen) + "\n")
    ylen = [len(v) for v in tmpMxDiffRG]
    # print ("ylen:", ylen)
    writing_file2.write("ylen: " + str(ylen) + "\n")
    zoneBreadth = ylen[0] // nDivide
#    """
    for x in range(xlen):
        if x == int(xlen / 2):
#        for y in range(ylen[0]):
            for yi in range(nDivide):
                sum = 0
                n = 0
                for y in range(yi * zoneBreadth, (yi + 1) * zoneBreadth):
                    sum = sum + tmpMxDiffRG[x][y]
                    n = n + 1
                sum = sum / n
                writing_file2.write(str(yi) + ',' + str(sum) + "\n")
#    """
    writing_file2.close()

#��---------------------------------
#### Calculating correcting value
def calculate_correct_value(tmpMxDiffRG, nYDivide, xStart, xEnd, yStart, yEnd, height2, mean2):
    """
      height2: height of image (unit: pixels)
      mean2[n of blocks]: mean R-G dif value in blocks
      tmpMxDiffRG[width pixes] x [n of blocks]: R-G difference value in ROI 
      nYDivede: number of divided block for calculation
      zoneLUT[height2]: LUT of attributed block number of each row
      xStart, xEnd, yStart, yEnd: coordinate of ROI
    """
    maxcorrectValue = 0
    maxCorV = [0.0 for yz in range(nYDivide)]
    correctValueMx = [[0.0 for yz in range(nYDivide)] for xi in range(xEnd - xStart)]
    zoneBreadth = height2 // nYDivide # height of a Y zone
    print ("zoneBreadth: ", zoneBreadth)
    zoneLUT = [0] * height2 # set look up table for y -> yzone
    for y in range(yStart, yStart + height2):
        zoneLUT[y] = min (y // zoneBreadth, nYDivide - 1) # make LUT for y -> yzone
    timenow = datetime.datetime.now()
    print("Start Calculate correctValueMx:", timenow)
#### Calculating correcting value
    for xi in range(xEnd - xStart): # process in each x colums in ROI
        for yz in range(nYDivide):
            sumtmpDiffRG = 0.0
            nr = 0.0
            if (xi == 0) and (yz == int(nYDivide /2)):
                timenow = datetime.datetime.now()
                print("Calcurate ", xi, "th correctValueMx:", timenow)
            for xi2 in range(xi-2, xi+3): # calculating back and forth 5 rows
                for yj in range(yStart + yz * zoneBreadth, yStart + (yz + 1) * zoneBreadth): 
                    if (xi2 > 0) and (yj > 0) and (xi2 < (xEnd - xStart)) and (yj < yStart + height2): 
                       # summate R-G value difference in each Y zone
                        nr = nr + 1.0
                        sumtmpDiffRG = sumtmpDiffRG + tmpMxDiffRG[xi2][yj - yStart] #
            if nr > 0.0:
                correctValueMx[xi][yz] = (sumtmpDiffRG / nr) - mean2[yz]
#                correctValueMx[xi][yz] = abs(sumtmpDiffRG / nr) - abs(mean2[yz])  # 2022.5.27 modified not good
#       define correcting value
#           divide sum of R-G diff with pixel number and subtract normal R-G diff mean
            else: 
                correctValueMx[xi][yz] = 0.0
            if correctValueMx[xi][yz] < 0.0: # if R mean < G mean in a Y zone then no correction
                correctValueMx[xi][yz] = 0.0 
            if (xi % 100 == 0) and (yz == 5):
                 print ("xi: ", xi, "yz: ", yz, "correctValueMx: ", correctValueMx[xi][yz])
            maxcorrectValue = max(maxcorrectValue, correctValueMx[xi][yz]) # calculate max correction value
            maxCorV[yz] = max(maxCorV[yz], correctValueMx[xi][yz]) # calculate max correction value
    print ("maxcorrectValue", maxcorrectValue)
    return maxcorrectValue, correctValueMx, zoneLUT, maxCorV
    """
        max correctValue:
        correctValueMx[width pixels of ROI][n of blocks]:
        zoneLUT[height pixels of ROI]:
        maxCorV[n of blocks]:
    """
#-----------
# image correction processing log writing out
def image_correct_log_writeout(x, y, height2, yStart, Pcc, HPcc, MPcc, LPcc):
    if (x % 100 == 0) and (y == int(height2)/5 + yStart):
        HPcc = Pcc
    elif (x % 100 == 0) and (y == int(height2)/2 + yStart):
        MPcc = Pcc
    elif (x % 100 == 0) and (y == 4 * int(height2)/5 + yStart):
        LPcc = Pcc
        print("x(",x,")  G Mixing Ratio L:", LPcc, "M: ",MPcc, "H: ",HPcc)
    return  HPcc, MPcc, LPcc

######### Image correcting routine ################
def image_correct(impR, impR2, impG, xStart, xEnd, Mode, GmixRatio, G_gamma, nYDivide, mean2, yStart, yEnd, folder, mixGamma): 
# impR: R Image impR2: R image to be corrected mean: mean differance of R-G
# xStart/xEnd: x start/end value of ROI, Mode: bit depth of image: GmixRatio: Maximum R Ch. mixing ratio
# G_gamma: Gamma Value of G ch./  nYDivide: dividing number of ROI in Y direction
# mean2: mean R-G differance value in each divided ROIs 
    nYDivide = int(nYDivide) # number of Y zone
    width = impR.getWidth() # get image width
    if xEnd > width:
        xEnd = width
    height = impG.getHeight() # get image height
    if yStart == 0 and yEnd == 0:
        height2 = height
    else:
        height2 = yEnd - yStart
    correctValue = []
#### Getting R-G value difference in each pixel
    tmpMxDiffRG = get_RG_v_diff(impR, impG, xStart, xEnd, yStart, yEnd, height2, folder, nYDivide)
#### Calculating correcting value
    maxcorrectValue, correctValueMx, zoneLUT, maxCorV = calculate_correct_value(tmpMxDiffRG, nYDivide, xStart, xEnd, yStart, yEnd, height2, mean2)

### value correction process
    timenow = datetime.datetime.now()
    print("Start Correction:", timenow)
    HPcc = 0.0
    MPcc = 0.0
    LPcc = 0.0
    for x in range(xStart, xEnd): # in each x row in ROI
        for y in range(yStart, yStart + height2): # start R value correction
            PxR= impR.getPixel(x, y)
            PxG= impG.getPixel(x, y)
#       Darken or Lighten Mixing G ch. Value/ gamma=G_gamma
            if G_gamma != 1.0:
                PxGF = float(PxG[0])/ float(convValue(256, Mode) -1)
                PxG[0]= convValue(int(PxGF ** G_gamma * 255), Mode)

            if correctValueMx[x - xStart][zoneLUT[y - yStart]] < 0.0: # if correction value < 0 then no correction
                correctValueMx[x - xStart][zoneLUT[y - yStart]] = 0.0
            Pcc0 = abs(correctValueMx[x - xStart][zoneLUT[y - yStart]]) / maxcorrectValue
            """
              Pcc0: Correcting Ratio of the Pixel to Max Correction Value
            """
#             calculate correction value ratio
#            Pcc1 = Pcc0 ** (1/2.0) # gamma weighted
            """
            Pcc1 = Pcc0 ** mixGamma # gamma weighted < mix Gamma (2022.5.27)
            Pcc = Pcc1 * GmixRatio # GmixRatio: Max G ch. mixing Ratio (like..            """
            Pcc1 = Pcc0 * GmixRatio # gamma weighted < mix Gamma (2022.5.27)
            Pcc = Pcc1 ** mixGamma # GmixRatio: Max G ch. mixing Ratio (like... 0.7 or 0.9)
            """
              Pcc: Actual G ch. Mixing rate for the Pixel
            """
#             define corrected pixel value in R
            Pxc = (PxG[0] * Pcc + PxR[0] * (1-Pcc)) # Mixing G Value and R Value with ratio Pcc
            """
              Pxc: Actual Corrected R ch. Value of the Pixel
            """
            # write out processing log 
            HPcc, MPcc, LPcc = image_correct_log_writeout(x, y, height2, yStart, Pcc, HPcc, MPcc, LPcc)
            # set Pxc to R ch. / Pxc: corrected value for R ch.
            if Pxc <= 0:
                Pxc = 0
            elif Pxc >= (256 * convValue(1, Mode) - 1):
                Pxc = (256 * convValue(1, Mode) - 1)
            impR2.getProcessor().set(x, y, int(Pxc))

#----------------------------------------------
def manipulation(imps, Mode, Show, RegionSE, folder, filename, GmixRatio, G_gamma2, nYDivide, mixGamma):
# GmixRatio: Maximum G ch. value mixing ratio to R ch.
# G_gamma2: Gamma correction value for G ch. (from parameter input dialog)
# nYDivide: n of calcurating blocks (from parameter input dialog)
    """
      RegionSE[x][y] coordinates of ROI and Sampling Region
                 [y]: 0: ROI xStart, 1: ROI xEnd, 2: ROI yStart, 3: ROI yEnd
                      4: sampling Region xStart 5: sampling Region xEnd
              [x]     0: 1st ROI(& Sampling Region), 1: 2nd ROI, 2: 3rd ROI
    """
    imp0 = imps[0] #imp0: Original Image 
    if Mode == "Auto":
        bitdepth = imp0.getBitDepth()
        if bitdepth == 8:
            Mode = "8-bit"
        elif bitdepth ==16:
            Mode = "16-bit"
        elif bitdepth ==32:
            Mode = "32-bit"
        else:
            Mode = "16-bit"
    if Show == "Yes":
        imp0.show()
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    n = 0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
        ip = i.getProcessor()
        n = n + 1
    impR =  impsCh[0]
    impG =  impsCh[1]
    impB =  impsCh[2]
    if Mode <> "8bit":
        IJ.saveAsTiff(impG, folder + filename +"_G.tif")
        IJ.saveAsTiff(impB, folder + filename +"_B.tif")
    if Show == "Yes":
        impR.show()
        impG.show()
        impB.show()
    impR2 = Duplicator().run(impR) # impR2: correcting R image
    width = impR.getWidth()
    iwidth = int(width * 0.015) # define width (15%) of sampling region for calculating mean R-G difference

    i = 0
    while i < 3:
      #  IJ.showMessage(str(i) + "th routine " + str(RegionSE[i][1]) + " " + str(RegionSE[i][0]))
        if RegionSE[i][1] - RegionSE[i][0] >= 1:
            if RegionSE[i][5] - RegionSE[i][4] > 0:   #RegionSE[i][4]: SmplXEnd - RegionSE[i][5]: SmplXStart
                mean2 = calcMean(impR, impG, RegionSE[i][4], RegionSE[i][5], nYDivide, RegionSE[i][2], RegionSE[i][3], folder)
                                                           # RegionSE[i][2], RegionSE[i][3]: yStart, yEnd
            else:
                mean2 = calcMean(impR, impG, RegionSE[i][0]-iwidth, RegionSE[i][0]-1, nYDivide, RegionSE[i][2], RegionSE[i][3], folder)
          #  IJ.showMessage(str(i) + "th correcting routine")
            image_correct(impR, impR2, impG,  RegionSE[i][0], RegionSE[i][1], Mode, GmixRatio, G_gamma2[i], nYDivide, mean2, RegionSE[i][2], RegionSE[i][3], folder, mixGamma)
            i = i + 1
        else:
            break


    impR2.show()
    mergeimp = RGBStackMerge.mergeChannels([impR2, impG, impB, None, None, None, None], True)
#    IJ.saveAsTiff(mergeimp, folder + filename +"_Merge_Stack.tif")
    # convert the composite image to the RGB image
    RGBStackConverter.convertToRGB(mergeimp)
    IJ.saveAsTiff(mergeimp, folder + filename +"_Merge_8bit.tif")
    mergeimp.show()
    if Mode <> "8-bit":
        IJ.saveAsTiff(impR2, folder + filename +"_R2.tif")

    if Show == "Yes":
        imp0.hide()
        impR.hide()
        impR2.hide()
        impG.hide()
        impB.hide()
#        mergeimp.hide()
    imp0.close()
    impR.close()
    impR2.close()
    impG.close()
    impB.close()
#    mergeimp.close()

# ---ROI dialog-------
class RoiPreviewer(ItemListener, TextListener):
    def __init__(self, imp, ckbox1, nfield_xStart, nfield_xEnd, nfield_yStart, nfield_yEnd):
        """
             imp: an ImagePlus
             preview_checkbox: a java.awt.Checkbox controlling whether to
                                                 dynamically update the ImagePlus as the
                                                 scrollbar is updated, or not.
        """
        self.imp = imp
        self.original_ip = imp.getProcessor().duplicate() # store a copy
        self.nfield_xStart = nfield_xStart
        self.nfield_xEnd = nfield_xEnd
        self.nfield_yStart = nfield_yStart
        self.nfield_yEnd = nfield_yEnd
        self.prv_ckBox1 = ckbox1
    

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        if event.getStateChange() == event.SELECTED:
            self.changeRoi()

    def textValueChanged(self, event):
        if self.prv_ckBox1.getState():
            try:
                rx = Integer.parseInt(self.nfield_xStart.getText())
                ry = Integer.parseInt(self.nfield_yStart.getText())
                rxe = Integer.parseInt(self.nfield_xEnd.getText())
                rye = Integer.parseInt(self.nfield_yEnd.getText())
                self.changeRoi()
            except:
                return

    def reset(self):
        """ Restore the original ImageProcessor """
        self.imp.setProcessor(self.original_ip)
    
    def changeRoi(self):
        """ Execute the in-place scaling of the ImagePlus. """
        rx = Integer.parseInt(self.nfield_xStart.getText())
        ry = Integer.parseInt(self.nfield_yStart.getText())
        rxe = Integer.parseInt(self.nfield_xEnd.getText())
        rye = Integer.parseInt(self.nfield_yEnd.getText())
        rw = rxe - rx
        rh = rye - ry
        rx = int(rx)
        ry = int(ry)
        rw = int(rw)
        rh = int(rh)
        if rx + rw > self.imp.width:
           rw = self.imp.width - rx
        if ry + rh > self.imp.height:
           rh = self.imp.height - ry
        print (rx, ry, rw, rh)
        self.imp.setRoi(rx, ry, rw, rh)

def setRoiUI(imp, i, Title, Message, yStart, yEnd, Mode):
#    imp = WM.getCurrentImage()
    impw = imp.width
    imph = imp.height
    roi = imp.getRoi()
    if roi:
        rOrgx = roi.getBounds().x
        rOrgy = roi.getBounds().y
        rOrgw = roi.getBounds().width
        rOrgh = roi.getBounds().height
    else:
        rOrgx = 0
        rOrgy = 0
        rOrgw = 0
        rOrgh = 0
    
    if not imp:
        print "Open an image first!"
        Mode = "Stop"
        return
    gd = GenericDialog(Title)
    gd.addMessage(Message + "  Image Pixel Size: " + str(impw) + " x " + str(imph))
    gd.addNumericField("Start Point (x)", rOrgx, 0)
    gd.addNumericField("End Point (x)", rOrgx + rOrgw, 0)
    gd.addMessage("Y Region to be corrected")
    gd.addNumericField("Start Point (y)", yStart, 0)
    gd.addNumericField("End Point (y)", yEnd, 0)
    gd.addCheckbox("Preview ROI", True)
    gd.addMessage("If you would like to renew preview, \n clear checkbox once and check it again.")
    if Title != "Select Sampling Region":
        gd.addNumericField("Gamma Correction Value for G (darken > 1.0 / Lighten <1.0)", 1.00, 2)
        Choice1 = ["Yes","No"]
        gd.addChoice("Define Sampling Normal Region ?", Choice1, Choice1[1])
        Choice2 = ["Yes","No"]
        gd.addChoice("Define Next ROI?", Choice2, Choice2[1])

    # The UI elements for the above two inputs
    nfield_xStart = gd.getNumericFields().get(0) # 1st start
    nfield_xEnd = gd.getNumericFields().get(1) # 1st end
    nfield_yStart = gd.getNumericFields().get(2) # 1st start
    nfield_yEnd = gd.getNumericFields().get(3) # 1st end

    checkbox1 = gd.getCheckboxes().get(0)

    previewer = RoiPreviewer(imp, checkbox1, nfield_xStart, nfield_xEnd, nfield_yStart, nfield_yEnd)
    checkbox1.addItemListener(previewer)
    nfield_xStart.addTextListener(previewer)
    nfield_xEnd.addTextListener(previewer)
    nfield_yStart.addTextListener(previewer)
    nfield_yEnd.addTextListener(previewer)

    gd.showDialog()

    if gd.wasCanceled():
        previewer.reset()
        Mode   = "Stop"
        IJ.log("User Canceled")
        checkbox1.removeItemListener(previewer)
    else:
        previewer.changeRoi()
        checkbox1.removeItemListener(previewer)
    xStart =  gd.getNextNumber()
    xEnd =  gd.getNextNumber()
#    IJ.showMessage("xStart xEnd: " + str(xStart) + str(xEnd))
    if Title != "Select Sampling Region":
        yStart =  gd.getNextNumber()
        yEnd =  gd.getNextNumber()
#        IJ.showMessage("yStart yEnd: " + str(yStart) + str(yEnd))
        G_gamma = gd.getNextNumber()
        if G_gamma < 0.0:
            G_gamma = 0.1
#        IJ.showMessage("G_gamma" + str(G_gamma))
        smplYes = gd.getNextChoice()
        Next = gd.getNextChoice()
    if Mode == "Stop":
        impclose = IJ.getImage()
        impclose.hide()
        impclose.close()
    if Title != "Select Sampling Region":
        return Mode, int(xStart), int(xEnd), int(yStart), int(yEnd), smplYes, Next, G_gamma
    else:
        return Mode, int(xStart), int(xEnd)

#--parameter dialog 
def getParameters_0():
    listRSE = [[0 for i in range(2)] for j in range(3)]
    gd = GenericDialog("Light Fog Correction Tool: (c)2021 Yasuo Ohnishi")
    gd.addNumericField("Max Mixing Ratio of G Value (0.0 to 1.0)", 1.00, 2)
    gd.addNumericField("Power for mixing adjustment", 0.18, 2)
    gd.addMessage("Vertical Dividing Zone")
    gd.addNumericField("Number of Y Dividing Zone", 25, 0)
    Show = ["Yes","No"]
    gd.addChoice("Display working windows", Show, Show[0])
    gd.showDialog()
#
    parm_Mode   = "Auto"
    GmixRatio = gd.getNextNumber()
    mixGamma  = gd.getNextNumber()
    nYDivide = gd.getNextNumber()
    parm_Show   = gd.getNextChoice()
    """
    if GmixRatio > 1.0:  # Restrict Maximum G ch. Mixing Ratio
        GmixRatio = 1.0
    """
#
    if gd.wasCanceled():  
        parm_Mode   = "Stop"
        impclose = IJ.getImage()
        impclose.hide()
        impclose.close()
        IJ.log("Canceled")
#
    return str(parm_Mode), str(parm_Show), GmixRatio, nYDivide, mixGamma


## Main Program
RegionSE = [[0 for j in range(6)] for i in range(3)] # variable for ROI
meanIntensity = []
G_gamma2 = [0.0 for i in range(3)]
window = WM.getActiveWindow()
Mode = "Auto"
if not window:
    od = OpenDialog("Choose a file")
    folder = od.getDirectory()
    filename = od.getFileName()
    if filename:
        path = folder + filename
        IJ.log(path)
        IJ.log(folder)
        IJ.log(filename)
# Always Open file with Bio-format Plugin
        imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
        imp0 = imps[0] #imp0: Original Image 
        imp0.show()
    else:
        Mode ="Stop"
else:
    imp0 = IJ.getImage()
    filename = imp0.getTitle()
#    folder = IJ.getDirectory("image") # get image's original directory
#    folder = IJ.getDirectory("Output_directory")
    folder = imp0.getOriginalFileInfo().directory  # get image's original directory
    # IJ.showMessage(folder)
    path = folder + filename
if Mode == "Auto":
    bitdepth = imp0.getBitDepth()
    if bitdepth == 8:
        Mode = "8-bit"
    elif bitdepth ==16:
        Mode = "16-bit"
    elif bitdepth ==32:
        Mode = "32-bit"
    else:
        Mode = "16-bit"
Mode = str(Mode)

#  input Parameters 0
parm = getParameters_0()
if parm is not None:
# if parm is not null object, then...
#    Mode, Show, GmixRatio, G_gamma, nYDivide = parm
    Mode, Show, GmixRatio, nYDivide, mixGamma = parm
#    ------------
if Mode != "Stop":
    
    impWidth = imp0.width
    impHeight = imp0.height
    imp0.show
    i = 0
    while i < 3:
    #  input Parameters 1
    #    parm = getParameters_1(i, impWidth, impHeight)
        Title = "Select Correcting Region"
        Message = str(i + 1) + "th ROI Selection"
        WM.getActiveWindow().toFront()
        parm = setRoiUI(imp0, i, Title, Message, 0, impHeight, Mode)
        if parm[0] == "Stop":
            Mode = parm[0]
            break
        else:
    # if parm is not null object, then...
            Mode, RegionSE[i][0], RegionSE[i][1], RegionSE[i][2], RegionSE[i][3], smplYes, Next, G_gamma2[i]= parm
            """
            RegionSE[0][0]: xStart, RegionSE[0][1]: xEnd RegionSE[0][2]: yStart, RegionSE[0][3]: yEnd
            smplYes: Input Sampling Region, Next: Input Next Region
            """
    #    IJ.showMessage("smplYes = ",smplYes)
            if smplYes == "Yes":
        #        parm2 = getParameters_2(i)
                Title = "Select Sampling Region"
                Message = str(i + 1) + "th Sampling Region"
                WM.getActiveWindow().toFront()
                parm2 = setRoiUI(imp0, i, Title, Message, RegionSE[i][2], RegionSE[i][3], Mode)

        #  input Parameters
                if parm2[0] != "Stop":
                # if parm is not null object, then...
                    Mode, RegionSE[i][4], RegionSE[i][5] = parm2
                    """
                    RegionSE[0][4]: sampling region xStart, RegionSE[0][5]: sampling region xEnd
                    """
                else:
                    Mode = parm2[0]
                    break
            i = i + 1
            if Next == "No":
                break
#  IJ.showMessage(Mode)
if Mode != "Stop":
    manipulation(imps, Mode, Show, RegionSE, folder, filename, GmixRatio, G_gamma2, nYDivide, mixGamma)
IJ.log("Finish")
#IJ.run("Quit")

